--[[
CollisionDeactivater

Specialization to deactivate the collision of the bags at an certain fill level

Author: 	Ifko[nator]
Date: 		13.06.2021
Version:	1.3

History:	v1.0 @18.07.2020 - initial implemation in FS 19
			-------------------------------------------------------------------------------------------
			v1.1 @28.08.2020 - collision will now also adjust at loading, for pallets, they aren't full
			-------------------------------------------------------------------------------------------
			v1.2 @12.09.2020 - fix for the store icon generator
			-------------------------------------------------------------------------------------------
			v1.3 @13.06.2021 - minior adjustments
]]

CollisionDeactivater = {};

function CollisionDeactivater.prerequisitesPresent(specializations)
	return true;
end;

function CollisionDeactivater.registerEventListeners(vehicleType)
	local functionNames = {
		"onLoad",
		"onPostLoad",
		"onUpdate"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, CollisionDeactivater);
	end;
end;

function CollisionDeactivater:onLoad(savegame)
	if RealisticSeederUtil == nil then
		return;
	end;
	
	local specCollisionDeactivater = RealisticSeederUtil.getSpecByName(self, "collisionDeactivater");
	
	specCollisionDeactivater.oldFillLevel = 0;
	specCollisionDeactivater.collisions = {};
	specCollisionDeactivater.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.collisionDeactivaters#fillUnitIndex"), 1);
	specCollisionDeactivater.newCollisionMask = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.collisionDeactivaters#newCollisionMask"), 0);

	local collisionNumber = 0;

	while true do
		local collisionKey = "vehicle.collisionDeactivaters.collisionDeactivater(" .. tostring(collisionNumber) .. ")";

		if not hasXMLProperty(self.xmlFile, collisionKey) then
			break;
		end;

		local node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, collisionKey .. "#node"), self.i3dMappings);

		if node ~= nil then
			local fillPercent = getXMLFloat(self.xmlFile, collisionKey .. "#fillPercent");
			
			if fillPercent ~= nil then
				local collision = {};

				collision.node = node;
				collision.fillPercent = fillPercent;
				collision.startMask = getCollisionMask(collision.node);

				table.insert(specCollisionDeactivater.collisions, collision);
			end;
		end;

		collisionNumber = collisionNumber + 1;
	end;
end;

function CollisionDeactivater:onPostLoad(savegame)
	if RealisticSeederUtil == nil then
		return;
	end;
	
	local specCollisionDeactivater = RealisticSeederUtil.getSpecByName(self, "collisionDeactivater");
	
	if #specCollisionDeactivater.collisions > 0 then		
		for _, collision in ipairs(specCollisionDeactivater.collisions) do
			local currentFillPercentage = MathUtil.round(self:getFillUnitFillLevelPercentage(specCollisionDeactivater.fillUnitIndex), 3);

			if currentFillPercentage <= collision.fillPercent then
				setCollisionMask(collision.node, specCollisionDeactivater.newCollisionMask);
				
				specCollisionDeactivater.oldFillLevel = self:getFillUnitFillLevel(specCollisionDeactivater.fillUnitIndex);
			end;
		end;
	end;
end;

function CollisionDeactivater:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if RealisticSeederUtil == nil then
		return;
	end;

	local specCollisionDeactivater = RealisticSeederUtil.getSpecByName(self, "collisionDeactivater");
	
	if #specCollisionDeactivater.collisions > 0 then		
		if specCollisionDeactivater.oldFillLevel ~= self:getFillUnitFillLevel(specCollisionDeactivater.fillUnitIndex) then
			for _, collision in ipairs(specCollisionDeactivater.collisions) do
				local currentFillPercentage = MathUtil.round(self:getFillUnitFillLevelPercentage(specCollisionDeactivater.fillUnitIndex), 3);

				if currentFillPercentage <= collision.fillPercent then
					setCollisionMask(collision.node, specCollisionDeactivater.newCollisionMask);
				else
					setCollisionMask(collision.node, collision.startMask);
				end;
			end;

			specCollisionDeactivater.oldFillLevel = self:getFillUnitFillLevel(specCollisionDeactivater.fillUnitIndex);
		end;
	end;
end;