--[[
RealisticSeeder

Specialization for an realistic seeder

Author: 	Ifko[nator]
Date: 		31.07.2021
Version:	2.3

History:	v1.0 @04.07.2020 - initial implemation in FS 19
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.1 @30.07.2020 - changed the loading logic of the seed fillTypes
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.2 @16.08.2020 - fix for sugar cane planters and for additional tank fill volumes
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.3 @27.08.2020 - reworkt function to load fill types, so it is not mess up with other mods
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.4 @12.09.2020 - added support for trailers and shovels they use the fill type category 'bulk' or 'augerwagon'
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.5 @16.02.2021 - added possibility to unload seed in pallets from a trailer/shovel
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.6 @12.06.2021 - reworkt function to unload seed in pallets from a trailer/shovel
				             - added possibility to unload seed and fertilizer seperatly
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.7 @13.06.2021 - minior adjustments
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.8 @14.06.2021 - Fix following Error when try to join an multiplayer game:

			Error: Event initialization only allowed at compile time
			LUA call stack:
			  dataS/scripts/network/EventIds.lua (28) : printCallstack
			  =E:/Eigene Dokumente/My Games/FarmingSimulator2019/mods/FS19_realisticSeeder/specializations/RealisticSeeder.lua (1048) : InitStaticEventClass
			  =[C] (4294967295)
			  dataS/scripts/vehicles/SpecializationManager.lua (85) : source
			  dataS/scripts/main.lua (2631) : addSpecialization
			  =E:/Eigene Dokumente/My Games/FarmingSimulator2019/mods/FS19_realisticSeeder/specializations/RegisterSpecialization.lua (80) : addSpecialization
			  dataS/scripts/utils/Utils.lua (404) : newFunc
			  dataS/scripts/main.lua (2980) : loadVehicleTypeFromXML
			  dataS/scripts/gui/MPLoadingScreen.lua (677) : loadMod
			  dataS/scripts/misc/AsyncManager.lua (42) : lambda
			  dataS/scripts/misc/AsyncManager.lua (119) : runLambda
			  dataS/scripts/misc/DeferredLoadingManager.lua (39) : runTopTask
			  dataS/scripts/main.lua (1904) : update
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v1.9 @28.06.2021 - it's possible again to unload herbicide
							 - added possibility to switch between pallets and big bags for unload seed, fertilizer and lime
							 - on the hof bergmann map the default liquid fertilizer and herbicide tanks will replaced with the ibc tanks from the map for unloading
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v2.0 @09.07.2021 - fixed Multiplayer Errors
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v2.1 @13.07.2021 - added Function to create bags, if fill level is 100 liter or smaller
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v2.2 @24.07.2021 - it's possible again to unlaod to discharge triggers
			--------------------------------------------------------------------------------------------------------------------------------------------------------
			v2.3 @31.07.2021 - fix for the pronto 6 AS tank by Agrartechnik Nordeifel
]]

RealisticSeeder = {};
RealisticSeeder.currentModName = g_currentModName;
RealisticSeeder.XML_PATH_PREFIX = g_currentModDirectory .. "xml/";

function RealisticSeeder.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(FillUnit, specializations);
end;

function RealisticSeeder.registerEventListeners(vehicleType)
	local functionNames = {
		"onLoad",
		"onPostLoad",
		"onUpdate",
		"onUpdateTick",
		"onReadStream",
		"onWriteStream",
		"onReadUpdateStream",
		"onWriteUpdateStream",
		"onRegisterActionEvents"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, RealisticSeeder);
	end;
end;

function RealisticSeeder.registerFunctions(vehicleType)
	local newFunctions = {
		"loadSeedFillTypes",
		"getAttachedAdditionalTanks",
		"overwriteFillTypeMaterial",
		"addFillTypesToBulkAndAugerwagon",
		"unloadFillUnitsToPallets",
		"switchUnloadMode"
	};
	
	for _, newFunction in ipairs(newFunctions) do
		SpecializationUtil.registerFunction(vehicleType, newFunction, RealisticSeeder[newFunction]);
	end;
end;

function RealisticSeeder:onLoad(savegame)
	if RealisticSeederUtil == nil then
		return;
	end;

	RealisticSeeder.debugPriority = RealisticSeederUtil.getDebugPriority(RealisticSeederUtil.modDesc, "modDesc.realisticSeeder#debugPriority");

	if fileExists(RealisticSeederUtil.fillTypesPath) then
		self:addFillTypesToBulkAndAugerwagon(RealisticSeederUtil.fillTypesXML, self.xmlFile, getXMLString(self.xmlFile, "vehicle.storeData.category"), Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.base.size#width"), 3));
	else
		RealisticSeederUtil.printError("Can't load fillTypesXML file (path: " .. RealisticSeederUtil.fillTypesPath .. ")! Aborting adding fill types to teh fill type categories 'BULK' and 'AUGERWAGON'!", false, false, "RealisticSeeder");
	end;

	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

	specRealisticSeeder.hofBergmannNeedUpate = false;
	
	local _, _, bigBagModDir = RealisticSeederUtil.getModByTitle("LSFM - BigBag Pack");
	local _, _, hofBergmannModDir = RealisticSeederUtil.getModByTitle("Hof Bergmann");

	if hofBergmannModDir ~= "" then
		--## set filename to the big bags from the hof bergmann map
		bigBagModDir = hofBergmannModDir .. "objects/bigBags/";

		for _, nonSeedFillType in pairs({"FERTILIZER", "HERBICIDE"}) do
			local palletFilename = hofBergmannModDir .. "objects/ibcTanks/ibcTank_" .. string.lower(nonSeedFillType) .. ".xml";

			if nonSeedFillType == "FERTILIZER" then
				nonSeedFillType = "LIQUID" .. nonSeedFillType;
			end;

			local fillType = g_fillTypeManager:getFillTypeByName(nonSeedFillType);

			if fillType.palletFilename ~= nil 
				and fileExists(palletFilename) 
				and fillType.palletFilename ~= palletFilename 
			then
				--## switch default tanks for liquid fertilizer and herbicide with the ibc tanks from the hof bergmann map
				fillType.palletFilename = palletFilename;
			end;
		end;

		local hofBergmannModDesc = loadXMLFile("modDesc", hofBergmannModDir .. "modDesc.xml");

		if hofBergmannModDesc ~= nil then
			local currentHofBergmannVersion = getXMLString(hofBergmannModDesc, "modDesc.version");
			
			if string.find(currentHofBergmannVersion, "%s") then --## check for space character
				--## cut string before the space character (e.g.: the string "1.0.0.8 Final" convert to "1.0.0.8")
				currentHofBergmannVersion = string.sub(currentHofBergmannVersion, 1, string.find(currentHofBergmannVersion, "%s") - 1);
			end;

			specRealisticSeeder.hofBergmannNeedUpate = RealisticSeederUtil.getIsUnSupportetValue(RealisticSeederUtil.unSupportetHofBergmannVersions, currentHofBergmannVersion);
		end;

		delete(hofBergmannModDesc);
	end;
	
	specRealisticSeeder.bigBagModDir = bigBagModDir;
	specRealisticSeeder.bagUnloadTreshold = 500;

	if not RealisticSeederUtil.getIsValidStoreCategory(getXMLString(self.xmlFile, "vehicle.storeData.category")) then
		return;
	end;
	
	if RealisticSeederUtil.fillTypesPath ~= nil and fileExists(RealisticSeederUtil.fillTypesPath) then
		self:loadSeedFillTypes(RealisticSeederUtil.fillTypesXML, self.xmlFile, getXMLString(self.xmlFile, "vehicle.storeData.category"), Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.base.size#width"), 3));
	else
		RealisticSeederUtil.printError("Can't load fillTypesXML file (path: " .. RealisticSeederUtil.fillTypesPath .. ")! Aborting loding fill types for sowing machines!", false, false, "RealisticSeeder");
	end;
end;

function RealisticSeeder:onPostLoad(savegame)
	if RealisticSeederUtil == nil or not RealisticSeederUtil.getIsValidStoreCategory(getXMLString(self.xmlFile, "vehicle.storeData.category")) then
		return;
	end;
	
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
	local specFillVolume = RealisticSeederUtil.getSpecByName(self, "fillVolume");

	specRealisticSeeder.oldSeedFruitType = FillType.UNKNOWN;
	specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
	specRealisticSeeder.additionalTankFillLevel = 0;
	specRealisticSeeder.additionalTank = nil;
	specRealisticSeeder.hasStoppedWorkingAnimation = false;
	specRealisticSeeder.timerDuration = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sowingMachine#timerDuration"), 40);
	specRealisticSeeder.isUnloadingToPallets = false;

	self:switchUnloadMode(specRealisticSeeder.isUnloadingToPallets);

	if specSowingMachine ~= nil and g_fillTypeManager:getFillTypeByIndex(specFillUnit.fillUnits[specSowingMachine.fillUnitIndex].fillType).name == "SEED_POTATO" then
		g_animationManager:setFillType(specSowingMachine.animationNodes, FillType.SEEDS);
	end;
end;

function RealisticSeeder:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	if RealisticSeederUtil == nil then
		return;
	end;
	
	if self.isClient then
		local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

		self:clearActionEventsTable(specRealisticSeeder.actionEvents);
		
		if isActiveForInput then
			for _, input in pairs({
				"UNLOAD_SEED_BUTTON",
				"UNLOAD_FERTILIZER_BUTTON",
				"UNLOAD_LIME_BUTTON",
				"SELECT_UNLOAD_MODE"
			}) do
				local _, actionEventId = self:addActionEvent(
					specRealisticSeeder.actionEvents, --## actionEventsTable
					InputAction[input], --## inputAction
					self, --## target
					RealisticSeeder.actionEventUnloadFillUnit, --## callback
					false, --## triggerUp
					true, --## triggerDown
					false, --## triggerAlways
					true, --## startActive
					nil --## callbackState
				);

				g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_HIGH);
				g_inputBinding:setActionEventTextVisibility(actionEventId, true);
				g_inputBinding:setActionEventActive(actionEventId, true);
				g_inputBinding:setActionEventText(actionEventId, RealisticSeederUtil.getL10nText("input_" .. input));
			end;
		end;
	end;
end;

function RealisticSeeder:actionEventUnloadFillUnit(actionName, inputValue, callbackState, isAnalog)
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
	
	if actionName == "SELECT_UNLOAD_MODE" then
		self:switchUnloadMode(not specRealisticSeeder.isUnloadingToPallets);
	else	
		self:unloadFillUnitsToPallets(true, actionName == "UNLOAD_FERTILIZER_BUTTON", actionName == "UNLOAD_LIME_BUTTON");
	end;
end;

function RealisticSeeder:switchUnloadMode(isUnloadingToPallets, noEventSend)
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

	if isUnloadingToPallets ~= specRealisticSeeder.isUnloadingToPallets then
		specRealisticSeeder.isUnloadingToPallets = isUnloadingToPallets;

		RealisticSeederSwitchUnloadModeEvent.sendEvent(self, isUnloadingToPallets, noEventSend);
	end;
end;

function RealisticSeeder:overwriteFillTypeMaterial(specFillUnit, specFillVolume, fillUnitIndex, fillType)
	--## Hardcore material overwrite. The function 'self:setFillUnitForcedMaterialFillType()' has not reasonable worked..
	local fillTypeCheck = g_fillTypeManager:getFillTypeByIndex(fillType);
	
	if fillType == nil or fillTypeCheck == nil then
		fillType = FillType.SEEDS;
	end;

	local material = g_materialManager:getMaterial(fillType, "fillplane", 1);

	if material ~= nil then
		local fillPlane = specFillUnit.fillUnits[fillUnitIndex].fillPlane;
		
		if fillPlane ~= nil then
			for _, node in ipairs(fillPlane.nodes) do
				setMaterial(node.node, material, 0);
			end;
		end;
	
		if specFillVolume ~= nil then
			local mapping = specFillVolume.fillUnitFillVolumeMapping[fillUnitIndex];
			
			if mapping ~= nil then
				for _, volume in ipairs(mapping.fillVolumes) do
					setMaterial(volume.volume, material, 0);
				end;
			end;
		end;
	end;		
end;

function RealisticSeeder:addFillTypesToBulkAndAugerwagon(fillTypesXML, xmlFile, storeCategory, width)
	if fillTypesXML == nil then
		RealisticSeederUtil.printError("Missing fillTypesXML file! Aborting adding fill types for " .. self:getFullName() .. "!", false, false, "RealisticSeeder");

		return;
	elseif xmlFile == nil then
		RealisticSeederUtil.printError("Missing xmlFile file! Aborting adding fill types for " .. self:getFullName() .. "!", false, false, "RealisticSeeder");

		return;
	elseif storeCategory == nil then
		RealisticSeederUtil.printError("Missing storeCategory! Aborting adding types for " .. self:getFullName() .. "!", false, false, "RealisticSeeder");

		return;
	end;
	
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");

	local fillTypeNames = "";
	local fillTypeNumber = 0;

	while true do
		local fillTypeKey = "map.fillTypes.fillType(" .. tostring(fillTypeNumber) .. ")";

		if not hasXMLProperty(fillTypesXML, fillTypeKey) then
			break;
		end;

		local fillType = Utils.getNoNil(getXMLString(fillTypesXML, fillTypeKey .. "#name"), "");

		if fillType ~= "" then
			if fillTypeNames ~= "" then 
				fillType = " " .. fillType;
			end;

			fillTypeNames = fillTypeNames .. fillType;
		end;

		fillTypeNumber = fillTypeNumber + 1;
	end;

	if fillTypeNames == "" then
		return;
	end;

	local fillTypes = g_fillTypeManager:getFillTypesByNames(fillTypeNames, "Warning: '" .. self:getFullName() .. "' has invalid fillType '%s'.");
	
	local baseKey = string.format("vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration(%d).fillUnits", Utils.getNoNil(self.configurations["fillUnit"], 1) - 1);
	
	local fillUnitNumber = 0;

    while true do
		local fillUnitKey = string.format("%s.fillUnit(%d)", baseKey, fillUnitNumber);
		
        if not hasXMLProperty(xmlFile, fillUnitKey) then
            break;
		end;

		if not RealisticSeederUtil.getIsValidStoreCategory(storeCategory) and 
			(
				specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[FillType.SEEDS] 
				or specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[FillType.FERTILIZER] 
				or specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[FillType.LIQUID_FERTILIZER]
			)
		then
			--## create table for unloading
			specFillUnit.unloading = {};

			--## set new values for table!
			unloading = {};
			unloading.node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, fillUnitKey .. "#node")) or self.rootNode;
			unloading.width = width + 18;
			unloading.offset = StringUtil.getVectorNFromString(getXMLString(xmlFile, fillUnitKey .. "#offset") or "0 0 0", 3);

			--## insert new values in table!
			table.insert(specFillUnit.unloading, unloading);
		end;
		
		if fillTypes ~= nil then
			for _, fillType in pairs(fillTypes) do
				if RealisticSeederUtil.getCanFruitBePlanted(fillType) then
					local fillTypeCategories = Utils.getNoNil(getXMLString(xmlFile, fillUnitKey .. "#fillTypeCategories"), "");

					for _, category in pairs({"BULK", "AUGERWAGON"}) do
						if fillTypeCategories ~= "" and string.find(string.upper(fillTypeCategories), category) then
    	    			    if not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] then
								specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] = true;

								RealisticSeederUtil.printDebug("Insert fill type " .. g_fillTypeManager:getFillTypeByIndex(fillType).name .. " to " .. self:getFullName(), RealisticSeeder.debugPriority, true, "RealisticSeeder");
							end;
						end;
					end;
				end;
			end;
		end;

		fillUnitNumber = fillUnitNumber + 1;
	end;
end;

function RealisticSeeder:loadSeedFillTypes(fillTypesXML, xmlFile, storeCategory, width)
	if fillTypesXML == nil then
		RealisticSeederUtil.printError("Missing fillTypesXML file! Aborting loading seed fill types for " .. self:getFullName() .. "!", false, false, "RealisticSeeder");

		return;
	elseif xmlFile == nil then
		RealisticSeederUtil.printError("Missing xmlFile file! Aborting loading seed fill types for " .. self:getFullName() .. "!", false, false, "RealisticSeeder");

		return;
	elseif storeCategory == nil then
		RealisticSeederUtil.printError("Missing storeCategory! Aborting loading seed fill types for " .. self:getFullName() .. "!", false, false, "RealisticSeeder");

		return;
	end;

	RealisticSeederUtil.printDebug("Loading seed fill types for " .. self:getFullName() .. "!", RealisticSeeder.debugPriority, true, "RealisticSeeder");
	
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");

	local fillTypeNames = "";
	local fillTypeNumber = 0;

	while true do
		local fillTypeKey = "map.fillTypes.fillType(" .. tostring(fillTypeNumber) .. ")";

		if not hasXMLProperty(fillTypesXML, fillTypeKey) then
			break;
		end;

		local fillType = Utils.getNoNil(getXMLString(fillTypesXML, fillTypeKey .. "#name"), "");

		if fillType ~= "" then
			if fillTypeNames ~= "" then 
				fillType = " " .. fillType;
			end;

			fillTypeNames = fillTypeNames .. fillType;
		end;

		fillTypeNumber = fillTypeNumber + 1;
	end;

	if fillTypeNames == "" then
		RealisticSeederUtil.printError("Unable to load new seed fill types! Aborting loading seed fill types for " .. self:getFullName() .. "!", false, false, "RealisticSeeder");
		
		return;
	end;
	
	local fillTypes = g_fillTypeManager:getFillTypesByNames(fillTypeNames, "Warning: '" .. self:getFullName() .. "' has invalid fillType '%s'.");

	local baseKey = string.format("vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration(%d).fillUnits", Utils.getNoNil(self.configurations["fillUnit"], 1) - 1);

	local fillUnitNumber = 0;

    while true do
		local fillUnitKey = string.format("%s.fillUnit(%d)", baseKey, fillUnitNumber);
		
        if not hasXMLProperty(xmlFile, fillUnitKey) then
            break;
		end;

		local defaultFillType = Utils.getNoNil(getXMLString(xmlFile, fillUnitKey .. "#fillTypes"), "");

		--## fix unloading width for all sowing machines! Thanks to GIANTS for this issue!
		if string.find(string.lower(defaultFillType), "seeds") and RealisticSeederUtil.getIsValidStoreCategory(storeCategory) then
			--## clear GIANTS shit out of table or create it, if it is missing! e.g.: potato planters!
			specFillUnit.unloading = {};

			--## set new values for table!
			unloading = {};
			unloading.node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, fillUnitKey .. "#node")) or self.rootNode;
			unloading.width = width + 18;
			unloading.offset = StringUtil.getVectorNFromString(getXMLString(xmlFile, fillUnitKey .. "#offset") or "0 0 0", 3);

			--## insert new values in table!
			table.insert(specFillUnit.unloading, unloading);
		end;
		
		if fillTypes ~= nil then
			for _, fillType in pairs(fillTypes) do
				if RealisticSeederUtil.getCanFruitBePlanted(fillType) then
					local seedsFillType = g_fillTypeManager:getFillTypeByIndex(fillType);

					if specSowingMachine ~= nil and specSowingMachine.seeds ~= nil and string.find(string.lower(defaultFillType), "seeds") then	
						for _, fruitType in pairs(specSowingMachine.seeds) do
							local seedsFruitType = g_fillTypeManager:getFillTypeByIndex(g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType));

							if seedsFruitType ~= nil then
								--## check if the fruit type is avaiable
								if string.sub(seedsFillType.name, 6) == seedsFruitType.name and not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] then
									specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] = true;

									RealisticSeederUtil.printDebug("Insert fill type " .. seedsFillType.name .. " to " .. self:getFullName(), RealisticSeeder.debugPriority, true, "RealisticSeeder");
								end;
							end;
						end;
					else
						if RealisticSeederUtil.foundFunctionAdditionalTank(xmlFile) then
							local seedsFruitType = g_fruitTypeManager:getFruitTypeByName(string.sub(seedsFillType.name, 6));

							if seedsFruitType ~= nil then
								--## difference between the store categories
								if storeCategory == "seeders" or storeCategory == "planters" then
									if #specFillUnit.fillUnits == 1 then
										--## check if fertilizer and seeds is avaiable in the additional tank, if not, add it!
										for _, fillTypeToCheck in pairs({"FERTILIZER", "SEEDS"}) do
											if not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[FillType[fillTypeToCheck]] then
												specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[FillType[fillTypeToCheck]] = true;

												RealisticSeederUtil.printDebug("Insert fill type " .. g_fillTypeManager:getFillTypeByIndex(FillType[fillTypeToCheck]).name .. " to " .. self:getFullName(), RealisticSeeder.debugPriority, true, "RealisticSeeder");
											end;
										end;
									end;

									for _, category in pairs({"SOWINGMACHINE", "PLANTER"}) do
										local categoryIndex = g_fruitTypeManager.categories[category];
										local categoryFruitTypes = g_fruitTypeManager.categoryToFruitTypes[categoryIndex];

        								if categoryFruitTypes ~= nil then
        								    for _, fruitType in ipairs(categoryFruitTypes) do
												if string.sub(seedsFillType.name, 6) == g_fillTypeManager:getFillTypeByIndex(g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType)).name and not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] then
													specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] = true;

													RealisticSeederUtil.printDebug("Insert fill type " .. seedsFillType.name .. " to " .. self:getFullName(), RealisticSeeder.debugPriority, true, "RealisticSeeder");
												end;
        								    end;
										end;
									end;
								else
									--## in case, someone make an additional tank for potato... Unreal, but who knows..
									if string.sub(seedsFillType.name, 6) == "POTATO" and not specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] then
										specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes[fillType] = true;

										RealisticSeederUtil.printDebug("Insert fill type " .. seedsFillType.name .. " to " .. self:getFullName(), RealisticSeeder.debugPriority, true, "RealisticSeeder");
									end;
								end;
							end;
						end;
					end;
				end;
			end;
		end;

		fillUnitNumber = fillUnitNumber + 1;
	end;
end;

RealisticSeeder.addFillUnitFillLevelBackup = SowingMachine.addFillUnitFillLevel; --## backup function for all sowingmachines they are not supportet by the realistic seeder

function RealisticSeeder:addFillUnitFillLevel(superFunc, farmId, fillUnitIndex, fillLevelDelta, fillType, toolType, fillInfo)
	if not RealisticSeederUtil.getIsValidStoreCategory(getXMLString(self.xmlFile, "vehicle.storeData.category")) then
		return RealisticSeeder.addFillUnitFillLevelBackup(self, superFunc, farmId, fillUnitIndex, fillLevelDelta, fillType, toolType, fillInfo);
	end;
	
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
	local specFillVolume = RealisticSeederUtil.getSpecByName(self, "fillVolume");

	if self:getFillUnitFillType(fillUnitIndex) ~= FillType.UNKNOWN and self:getFillUnitFillType(fillUnitIndex) ~= fillType then
		if specFillUnit.fillUnits[fillUnitIndex].fillLevel / math.max(specFillUnit.fillUnits[fillUnitIndex].capacity, 0.0001) > self:getFillTypeChangeThreshold() then
			return 0;
		end;
	end;

	local fillTypeDesc = g_fillTypeManager:getFillTypeByIndex(fillType);

	if fillTypeDesc ~= nil then
		if fillUnitIndex == specSowingMachine.fillUnitIndex then
			if string.sub(fillTypeDesc.name, 1, 5) == "SEED_" then
				if fillTypeDesc.name ~= "SEED_POTATO" then	
					self:overwriteFillTypeMaterial(specFillUnit, specFillVolume, fillUnitIndex);
				end;
			else
				--## force material of fillVolume to be seeds, if fillType is accepted
				if self:getFillUnitSupportsFillType(fillUnitIndex, fillType) then
					self:overwriteFillTypeMaterial(specFillUnit, specFillVolume, fillUnitIndex);
				end;

				--## switch from seeds to "fill type", if possible
				local fruitType = specSowingMachine.seeds[specSowingMachine.currentSeed];

				if fruitType ~= nil then
					local seedsFillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(fruitType);
				
					if seedsFillType ~= nil then
						if self:getFillUnitSupportsFillType(fillUnitIndex, seedsFillType) then
							self:overwriteFillTypeMaterial(specFillUnit, specFillVolume, fillUnitIndex, seedsFillType);
						end;
					end;
				end;
			end;
		end;
	end;

    return superFunc(self, farmId, fillUnitIndex, fillLevelDelta, fillType, toolType, fillInfo);
end;

SowingMachine.addFillUnitFillLevel = RealisticSeeder.addFillUnitFillLevel; --## need an hardcore overwrite... Otherwhise, the fill type will automaticlly set to "seeds"!

function RealisticSeeder:getAttachedAdditionalTanks(rootVehicle)
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
	
	if rootVehicle ~= nil then
		local specAttacherJoints = RealisticSeederUtil.getSpecByName(rootVehicle, "attacherJoints");

		if specAttacherJoints ~= nil then
			for _, implement in pairs(specAttacherJoints.attachedImplements) do
				local additionalTank = implement.object;
				local additionalTank_specAttachable = RealisticSeederUtil.getSpecByName(additionalTank, "attachable");

				if additionalTank_specAttachable.attacherVehicle ~= nil then
					local attacherJointVehicleSpec = RealisticSeederUtil.getSpecByName(additionalTank_specAttachable.attacherVehicle, "attacherJoints");
					
					if RealisticSeederUtil.foundFunctionAdditionalTank(additionalTank.xmlFile) then
						local additionalTank_specFillUnit = RealisticSeederUtil.getSpecByName(additionalTank, "fillUnit");
						local additionalTank_specFillVolume = RealisticSeederUtil.getSpecByName(additionalTank, "fillVolume");
						local seedFillUnitIndex = RealisticSeederUtil.getSeedFillUnitIndex(additionalTank.xmlFile, additionalTank, additionalTank_specFillUnit);

						if string.find(additionalTank.configFileName, "FS19_horschPronto6AS/pronto6AS_tank.xml") then
							--## hadcore fix for the pronto 6 AS tank by Agrartechnik Nordeifel.
							seedFillUnitIndex = 1;
						end;

						if specRealisticSeeder.additionalTankFillType == FillType.UNKNOWN then
							specRealisticSeeder.additionalTankFillType = additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillType;
						end;

						if string.find(g_fillTypeManager:getFillTypeByIndex(additionalTank:getFillUnitFillType(seedFillUnitIndex)).name, "SEED") then
 							specRealisticSeeder.additionalTank = additionalTank;
							specRealisticSeeder.additionalTankFillLevel = additionalTank_specFillUnit.fillUnits[seedFillUnitIndex].fillLevel;

							local seedsFillType = g_fillTypeManager:getFillTypeByName(string.sub(g_fillTypeManager:getFillTypeByIndex(additionalTank:getFillUnitFillType(seedFillUnitIndex)).name, 6));

							if seedsFillType ~= nil then
								additionalTank:setFillUnitFillTypeToDisplay(seedFillUnitIndex, seedsFillType.index, true)
							end;

							if specRealisticSeeder.additionalTankFillLevel <= 0 then
								specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
								specRealisticSeeder.additionalTank = nil;
							end;

							break;
						else
							specRealisticSeeder.additionalTank = nil;
							specRealisticSeeder.additionalTankFillLevel = 0;

							additionalTank:setFillUnitFillTypeToDisplay(seedFillUnitIndex, additionalTank:getFillUnitFillType(seedFillUnitIndex), true);
						end;
					else
						specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
						specRealisticSeeder.additionalTank = nil;
						specRealisticSeeder.additionalTankFillLevel = 0;
					end;

					if additionalTank.getAttachedAdditionalTanks ~= nil then
						additionalTank:getAttachedAdditionalTanks(additionalTank);
					end;
				end;
			end;
		end;
	else
		specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
		specRealisticSeeder.additionalTank = nil;
		specRealisticSeeder.additionalTankFillLevel = 0;
	end;
end;

function RealisticSeeder:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if RealisticSeederUtil == nil then
		return;
	end;
	
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

	if specRealisticSeeder.hofBergmannNeedUpate then
		if g_realisticSeederHasShownWarning == nil then	
			g_currentMission:showBlinkingWarning(RealisticSeederUtil.getL10nText("warning_pleaseUpdateHofBergmann"), 12000);

			g_realisticSeederHasShownWarning = true;
		end;
	end;

	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

	local isSeedFillType, isFertilizerFillType, isLimeFillType, _, _, seedFillTypeIndex = RealisticSeederUtil.getLoadedFillTypes(self, self:getIsActiveForInput(), specFillUnit, #specFillUnit.fillUnits);
	
	local actionEventUnloadSeed = specRealisticSeeder.actionEvents[InputAction.UNLOAD_SEED_BUTTON];
	local actionEventUnloadFertilizer = specRealisticSeeder.actionEvents[InputAction.UNLOAD_FERTILIZER_BUTTON];
	local actionEventUnloadLime = specRealisticSeeder.actionEvents[InputAction.UNLOAD_LIME_BUTTON];
	local actionEventSelectUnloadMode = specRealisticSeeder.actionEvents[InputAction.SELECT_UNLOAD_MODE];
	
	local selectedUnloadMode = "Bigbags";
	
	local allowSwitchUnloadMode = true;

	if specRealisticSeeder.isUnloadingToPallets then
		selectedUnloadMode = RealisticSeederUtil.getL10nText("category_pallets");
	end;

	if actionEventSelectUnloadMode ~= nil then
		if actionEventUnloadSeed ~= nil then
			if seedFillTypeIndex ~= 0 then
				local seedTitle = g_fillTypeManager:getFillTypeByIndex(seedFillTypeIndex).title;
				local seedName = g_fillTypeManager:getFillTypeByIndex(seedFillTypeIndex).name;
				local seedFillUnitIndex = RealisticSeederUtil.getSeedFillUnitIndex(self.xmlFile, self, specFillUnit);
				local fillLevel = self:getFillUnitFillLevel(seedFillUnitIndex);
				local selectedSeedUnloadMode = selectedUnloadMode;
				
				if seedName == "SEED_POTATO" then
					selectedSeedUnloadMode = RealisticSeederUtil.getL10nText("category_pallets");
					
					allowSwitchUnloadMode = false;
				else
					if fillLevel <= specRealisticSeeder.bagUnloadTreshold then
						selectedSeedUnloadMode = RealisticSeederUtil.getL10nText("action_bags");

						allowSwitchUnloadMode = false;
					end;
				end;

				g_inputBinding:setActionEventText(actionEventUnloadSeed.actionEventId, string.format(RealisticSeederUtil.getL10nText("action_unloadXToX"), seedTitle, selectedSeedUnloadMode));
			end;

			g_inputBinding:setActionEventActive(actionEventUnloadSeed.actionEventId, isSeedFillType);
		end;

		if actionEventUnloadFertilizer ~= nil then
			if isFertilizerFillType then
				local fertilizerTitle = g_fillTypeManager:getFillTypeByName("FERTILIZER").title;
				
				g_inputBinding:setActionEventText(actionEventUnloadFertilizer.actionEventId, string.format(RealisticSeederUtil.getL10nText("action_unloadXToX"), fertilizerTitle, selectedUnloadMode));
			end;

			g_inputBinding:setActionEventActive(actionEventUnloadFertilizer.actionEventId, isFertilizerFillType);
		end;

		if actionEventUnloadLime ~= nil then
			if isLimeFillType then
				local limeTitle = g_fillTypeManager:getFillTypeByName("LIME").title;

				g_inputBinding:setActionEventText(actionEventUnloadLime.actionEventId, string.format(RealisticSeederUtil.getL10nText("action_unloadXToX"), limeTitle, selectedUnloadMode));
			end;
		
			g_inputBinding:setActionEventActive(actionEventUnloadLime.actionEventId, isLimeFillType);
		end;

		g_inputBinding:setActionEventActive(actionEventSelectUnloadMode.actionEventId, (isSeedFillType and allowSwitchUnloadMode) or (isFertilizerFillType or isLimeFillType));
	end;
end;

function RealisticSeeder:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if RealisticSeederUtil == nil then
		return;
	end;

	if not RealisticSeederUtil.getIsValidStoreCategory(getXMLString(self.xmlFile, "vehicle.storeData.category")) then
		return;
	end;
	
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");

	self:getAttachedAdditionalTanks(self:getRootVehicle());

	if specRealisticSeeder.additionalTankFillLevel <= 0 then
		specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
	end;

	if specSowingMachine ~= nil then
		local selectedSeedFruitType = specFillUnit.fillUnits[specSowingMachine.fillUnitIndex].fillType;

		if specRealisticSeeder.additionalTankFillType ~= FillType.UNKNOWN then
			selectedSeedFruitType = specRealisticSeeder.additionalTankFillType;
		end;

		if selectedSeedFruitType ~= nil then
			specSowingMachine.allowsSeedChanging = selectedSeedFruitType == FillType.SEEDS; --## with default seeds, you can switch the seed fruit types as in the base game
			
			if selectedSeedFruitType ~= FillType.UNKNOWN and not specSowingMachine.allowsSeedChanging and specRealisticSeeder.oldSeedFruitType ~= selectedSeedFruitType then
				local fillTypeDesc = g_fillTypeManager:getFillTypeByIndex(selectedSeedFruitType);
				
				if fillTypeDesc ~= nil then
					local fruitTypeDesc = g_fruitTypeManager:getFruitTypeByName(string.sub(fillTypeDesc.name, 6));
					
					if fruitTypeDesc ~= nil then
						self:setSeedFruitType(fruitTypeDesc.index, true);

						specRealisticSeeder.oldSeedFruitType = selectedSeedFruitType;
					end;
				end;
			end;

			if self:getIsTurnedOn() then
				if self:getIsImplementChainLowered() then
					if specRealisticSeeder.hasStoppedWorkingAnimation then	
						if specRealisticSeeder.timer > 0 then
							specRealisticSeeder.timer = specRealisticSeeder.timer - 1;
						else
							g_animationManager:startAnimations(specSowingMachine.animationNodes);
							
							specRealisticSeeder.hasStoppedWorkingAnimation = false;
						end;
					end;
				else
					if not specRealisticSeeder.hasStoppedWorkingAnimation then
						g_animationManager:stopAnimations(specSowingMachine.animationNodes);
						
						specRealisticSeeder.hasStoppedWorkingAnimation = true;
						specRealisticSeeder.timer = specRealisticSeeder.timerDuration;
					end;
				end;
			end;

			if g_fillTypeManager:getFillTypeByIndex(selectedSeedFruitType).name == "SEED_POTATO" then
				g_animationManager:setFillType(specSowingMachine.animationNodes, FillType.SEEDS);
			end;
		end;
	end;
end;

function RealisticSeeder:onReadStream(streamId, connection)
	if RealisticSeederUtil == nil then
		return;
	end;
	
	if connection:getIsServer() then
		local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

		if specRealisticSeeder.additionalTankFillType == nil then	
			specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
		end;

		specRealisticSeeder.additionalTankFillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS);
	end;
end;

function RealisticSeeder:onWriteStream(streamId, connection)
	if RealisticSeederUtil == nil then
		return;
	end;
	
	if not connection:getIsServer() then
		local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

		if specRealisticSeeder.additionalTankFillType == nil then	
			specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
		end;	
			
		streamWriteUIntN(streamId, specRealisticSeeder.additionalTankFillType, FillTypeManager.SEND_NUM_BITS);
	end;
end;

function RealisticSeeder:onReadUpdateStream(streamId, timestamp, connection)
	if RealisticSeederUtil == nil then
		return;
	end;

	if connection:getIsServer() then
		local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

		if specRealisticSeeder.additionalTankFillType == nil then	
			specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
		end;

		specRealisticSeeder.additionalTankFillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS);
	end;
end;

function RealisticSeeder:onWriteUpdateStream(streamId, connection, dirtyMask)
	if RealisticSeederUtil == nil then
		return;
	end;
	
	if not connection:getIsServer() then
		local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");

		if specRealisticSeeder.additionalTankFillType == nil then	
			specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
		end;
		
		streamWriteUIntN(streamId, specRealisticSeeder.additionalTankFillType, FillTypeManager.SEND_NUM_BITS);
	end;
end;

function RealisticSeeder:onStartWorkAreaProcessing(superFunc, dt)
	if RealisticSeederUtil == nil then
		return superFunc(dt);
	end;
	
	local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
    local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	
	specSowingMachine.isWorking = false;
    specSowingMachine.isProcessing = false;
	
	local seedsFruitType = specSowingMachine.seeds[specSowingMachine.currentSeed];
    local dx, _, dz = localDirectionToWorld(specSowingMachine.directionNode, 0, 0, 1);
    local angleRad = MathUtil.getYRotationFromDirection(dx, dz);
    local desc = g_fruitTypeManager:getFruitTypeByIndex(seedsFruitType);
	
	if desc ~= nil and desc.directionSnapAngle ~= 0 then
        angleRad = math.floor(angleRad / desc.directionSnapAngle + 0.5) * desc.directionSnapAngle;
    end;
	
	local angle = FSDensityMapUtil.convertToDensityMapAngle(angleRad, g_currentMission.terrainDetailAngleMaxValue);
	local seedsVehicle, seedsVehicleFillUnitIndex;

	if specRealisticSeeder.additionalTankFillType == nil then	
		specRealisticSeeder.additionalTankFillType = FillType.UNKNOWN;
	end;

	
	
	if specRealisticSeeder.additionalTankFillType == FillType.UNKNOWN then
		if self:getFillUnitFillLevel(specSowingMachine.fillUnitIndex) > 0 then
    	    seedsVehicle = self;
			seedsVehicleFillUnitIndex = specSowingMachine.fillUnitIndex;
		end;
	else
		if specRealisticSeeder.additionalTank ~= nil then
			local fillType = specRealisticSeeder.additionalTankFillType;
			local seedFillUnitIndex = 1;
			
			if specRealisticSeeder.additionalTank.xmlFile ~= nil and specRealisticSeeder.additionalTank.spec_fillUnit ~= nil then
				--seedFillUnitIndex = RealisticSeederUtil.getSeedFillUnitIndex(specRealisticSeeder.additionalTank.xmlFile, specRealisticSeeder.additionalTank, specRealisticSeeder.additionalTank.spec_fillUnit);
			end;

			if specRealisticSeeder.additionalTank:getFillUnitFillLevel(seedFillUnitIndex) > 0 and specRealisticSeeder.additionalTank:getFillUnitFillType(seedFillUnitIndex) == fillType then
        		seedsVehicle = specRealisticSeeder.additionalTank;
				seedsVehicleFillUnitIndex = seedFillUnitIndex;
        	end;
		end;
	end;

	if seedsVehicle ~= nil and seedsVehicle ~= self then
		local fillType = g_fruitTypeManager:getFillTypeIndexByFruitTypeIndex(seedsFruitType);
		
        seedsVehicle:setFillUnitFillTypeToDisplay(seedsVehicleFillUnitIndex, fillType);
	end;
	
    --## check if selected fruit is available in current map
	local canFruitBePlanted = RealisticSeederUtil.getCanFruitBePlanted(nil, seedsFruitType);
	
    if specSowingMachine.showWrongFruitForMissionWarning then
        specSowingMachine.showWrongFruitForMissionWarning = false;
	end;
	
    specSowingMachine.showFruitCanNotBePlantedWarning = not canFruitBePlanted;
    specSowingMachine.workAreaParameters.isActive = not specSowingMachine.needsActivation or self:getIsTurnedOn();
    specSowingMachine.workAreaParameters.canFruitBePlanted = canFruitBePlanted;
    specSowingMachine.workAreaParameters.seedsFruitType = seedsFruitType;
    specSowingMachine.workAreaParameters.angle = angle;
    specSowingMachine.workAreaParameters.seedsVehicle = seedsVehicle;
    specSowingMachine.workAreaParameters.seedsVehicleFillUnitIndex = seedsVehicleFillUnitIndex;
    specSowingMachine.workAreaParameters.lastTotalArea = 0;
    specSowingMachine.workAreaParameters.lastChangedArea = 0;
    specSowingMachine.workAreaParameters.lastStatsArea = 0;
end;

SowingMachine.onStartWorkAreaProcessing = RealisticSeeder.onStartWorkAreaProcessing; --## need an hardcore overwrite... Otherwhise, the seed fill type will automaticlly set to "seeds"!

function RealisticSeeder:unloadFillUnitsToPallets(showWarning, unloadFertilizer, unloadLime)
	if not self.isServer then
		g_client:getServerConnection():sendEvent(RealisticSeederUnloadFillUnitsToPalletsEvent:new(self, unloadFertilizer, unloadLime));
	else
		local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
		local specRealisticSeeder = RealisticSeederUtil.getSpecByName(self, "realisticSeeder");
		local unloadingPlaces = specFillUnit.unloading;
		local places = {};

		for _, unloading in ipairs(unloadingPlaces) do
			local node = unloading.node;
			
			local ox, oy, oz = unpack(unloading.offset);
			local x, y, z = localToWorld(node, ox - unloading.width * 0.1, oy, oz);
			
			local place = {
				startZ = z,
				startY = y,
				startX = x
			};

			place.rotX, place.rotY, place.rotZ = getWorldRotation(node);
			place.dirX, place.dirY, place.dirZ = localDirectionToWorld(node, 1, 0, 0);
			place.dirPerpX, place.dirPerpY, place.dirPerpZ = localDirectionToWorld(node, 0, 0, 1);
			place.yOffset = 1;
			place.maxWidth = math.huge;
			place.maxLength = math.huge;
			place.width = unloading.width;

			table.insert(places, place);
		end;

		local usedPlaces = {};
		local success = true;
		local availablePallets = {};

		for fillUnitNumber, fillUnit in ipairs(self:getFillUnits()) do
			local fillLevel = self:getFillUnitFillLevel(fillUnitNumber);
			local fillTypeIndex = self:getFillUnitFillType(fillUnitNumber);
			local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex);
			local fillTypeName = "SEED";

			if unloadFertilizer then
				fillTypeName = "FERTILIZER";
			elseif unloadLime then
				fillTypeName = "LIME";
			end;	

			if fillType ~= nil and string.find(fillType.name, fillTypeName) then
				local fixedFillTypeName = RealisticSeederUtil.fixFillTypeName(string.lower(fillType.name)); --## fix invalid case warning's
				
				local palletPath = "pallets/";
				
				if fillLevel <= specRealisticSeeder.bagUnloadTreshold and fixedFillTypeName ~= "seed_potato" then
					palletPath = "bags/bag_";
				end;

				--## realistic seeder seed fruit types pallets
				local palletFilename = RealisticSeeder.XML_PATH_PREFIX .. palletPath .. fixedFillTypeName .. ".xml";

				if fixedFillTypeName == "seeds" and fillLevel > specRealisticSeeder.bagUnloadTreshold then
					--## default seeds pallet for default seeds
					palletFilename = "data/objects/pallets/pioneer/pioneerPallet.xml";
				end;
				
				if not specRealisticSeeder.isUnloadingToPallets and fixedFillTypeName ~= "seed_potato" and fillLevel > specRealisticSeeder.bagUnloadTreshold then
					--## realistic seeder seed fruit types big bags
					palletFilename = RealisticSeeder.XML_PATH_PREFIX .. "bigBags/bigBag_" .. fixedFillTypeName .. ".xml";

					if fixedFillTypeName == "seeds" then
						--## default seeds
						if specRealisticSeeder.bigBagModDir ~= "" then
							--## seeds big bag from the big bag pack or the hof bergmann map for default seeds
							palletFilename = specRealisticSeeder.bigBagModDir .. "bigBagSeeds.xml";
						else
							--## default seeds big bag, when the lsfm big bag pack or the hof bergmann map is not loaded
							palletFilename = "data/objects/bigBagContainer/bigBagContainerSeeds.xml";
						end;
					elseif fixedFillTypeName == "fertilizer" then
						--## fertilizer
						if specRealisticSeeder.bigBagModDir ~= "" then
							--## fertilizer big bag from the big bag pack or the hof bergmann map
							palletFilename = specRealisticSeeder.bigBagModDir .. "bigBagFertilizer.xml";
						else
							--## default fertilizer big bag, when the lsfm big bag pack or the hof bergmann map is not loaded
							palletFilename = "data/objects/bigBagContainer/bigBagContainerFertilizer.xml";
						end;
					elseif fixedFillTypeName == "lime" then
						--## lime
						if specRealisticSeeder.bigBagModDir ~= "" then
							if not specRealisticSeeder.hofBergmannNeedUpate then	
								--## lime big bag from the big bag pack or the hof bergmann map
								palletFilename = specRealisticSeeder.bigBagModDir .. "bigBagLime.xml";
							else
								--## default lime big bag, when the hof bergmann map need an update
								palletFilename = "data/objects/bigBagContainer/bigBagContainerLime.xml";
							end;
						else
							--## default lime big bag, when the lsfm big bag pack or the hof bergmann map is not loaded
							palletFilename = "data/objects/bigBagContainer/bigBagContainerLime.xml";
						end;
					end;
				end;
				
				if fillType.palletFilename ~= nil and fileExists(palletFilename) then
					--## switch pallet filename with big bag and vice versa
					fillType.palletFilename = palletFilename;
				end;

				if fillUnit.canBeUnloaded and fillLevel > 0 and fillType.palletFilename ~= nil then
					while fillLevel > 0 do
						local pallet = availablePallets[fillTypeIndex];
						local isNewPallet = false;

						if pallet == nil then
							local sizeWidth, sizeLength, widthOffset, lengthOffset = StoreItemUtil.getSizeValues(fillType.palletFilename, "vehicle", 0, {});
							local x, _, z, place, width, _ = PlacementUtil.getPlace(places, sizeWidth, sizeLength, widthOffset, lengthOffset, usedPlaces, true, true, true);

							if x == nil then
								success = false;

								break;
							end;

							PlacementUtil.markPlaceUsed(usedPlaces, place, width);

							pallet = g_currentMission:loadVehicle(fillType.palletFilename, x, nil, z, 0, place.rotY, true, 0, Vehicle.PROPERTY_STATE_OWNED, self:getOwnerFarmId(), nil, nil);

							pallet:emptyAllFillUnits(true);

							isNewPallet = true;
						end

						local fillUnitIndex = pallet:getFirstValidFillUnitToFill(fillTypeIndex);

						if fillUnitIndex ~= nil then
							local appliedDelta = pallet:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, fillLevel, fillTypeIndex, ToolType.UNDEFINED, nil);

							self:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitNumber, -appliedDelta, fillTypeIndex, ToolType.UNDEFINED, nil);

							fillLevel = fillLevel - appliedDelta;
						else
							if isNewPallet then
								break;
							end;

							availablePallets[fillTypeIndex] = nil;
						end;
					end;
				end;
			end;
		end;

		if showWarning and not success then
			g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_INFO, g_i18n:getText("fillUnit_unload_nospace"));
		end;

		return success;
	end;
end;

function RealisticSeeder:unloadFillUnits(superFunc, ignoreWarning)
	if not self.isServer then
		g_client:getServerConnection():sendEvent(FillUnitUnloadEvent:new(self));
	else
		local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
		local unloadingPlaces = specFillUnit.unloading;
		local places = {};

		for _, unloading in ipairs(unloadingPlaces) do
			local node = unloading.node;
			
			local ox, oy, oz = unpack(unloading.offset);
			local x, y, z = localToWorld(node, ox - unloading.width * 0.1, oy, oz);
			
			local place = {
				startZ = z,
				startY = y,
				startX = x
			};

			place.rotX, place.rotY, place.rotZ = getWorldRotation(node);
			place.dirX, place.dirY, place.dirZ = localDirectionToWorld(node, 1, 0, 0);
			place.dirPerpX, place.dirPerpY, place.dirPerpZ = localDirectionToWorld(node, 0, 0, 1);
			place.yOffset = 1;
			place.maxWidth = math.huge;
			place.maxLength = math.huge;
			place.width = unloading.width;

			table.insert(places, place);
		end;

		local usedPlaces = {};
		local success = true;
		local availablePallets = {};

		for fillUnitNumber, fillUnit in ipairs(self:getFillUnits()) do
			local fillLevel = self:getFillUnitFillLevel(fillUnitNumber);
			local fillTypeIndex = self:getFillUnitFillType(fillUnitNumber);
			local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex);

			if fillUnit.canBeUnloaded and fillLevel > 0 and fillType.palletFilename ~= nil then
				while fillLevel > 0 do
					local pallet = availablePallets[fillTypeIndex];
					local isNewPallet = false;

					if pallet == nil then
						local sizeWidth, sizeLength, widthOffset, lengthOffset = StoreItemUtil.getSizeValues(fillType.palletFilename, "vehicle", 0, {});
						local x, _, z, place, width, _ = PlacementUtil.getPlace(places, sizeWidth, sizeLength, widthOffset, lengthOffset, usedPlaces, true, true, true);

						if x == nil then
							success = false;

							break;
						end;

						PlacementUtil.markPlaceUsed(usedPlaces, place, width);

						pallet = g_currentMission:loadVehicle(fillType.palletFilename, x, nil, z, 0, place.rotY, true, 0, Vehicle.PROPERTY_STATE_OWNED, self:getOwnerFarmId(), nil, nil);

						pallet:emptyAllFillUnits(true);

						isNewPallet = true;
					end

					local fillUnitIndex = pallet:getFirstValidFillUnitToFill(fillTypeIndex);

					if fillUnitIndex ~= nil then
						local appliedDelta = pallet:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, fillLevel, fillTypeIndex, ToolType.UNDEFINED, nil);

						self:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitNumber, -appliedDelta, fillTypeIndex, ToolType.UNDEFINED, nil);

						fillLevel = fillLevel - appliedDelta;
					else
						if isNewPallet then
							break;
						end;

						availablePallets[fillTypeIndex] = nil;
					end;
				end;
			end;
		end;

		if (ignoreWarning == nil or not ignoreWarning) and not success then
			g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_INFO, g_i18n:getText("fillUnit_unload_nospace"));
		end;

		return success;
	end;
end;

FillUnit.unloadFillUnits = RealisticSeeder.unloadFillUnits;

function RealisticSeeder:updateUnloadActionDisplay(superFunc)
	if RealisticSeederUtil == nil then
		return superFunc();
	end;
	
	local specFillUnit = RealisticSeederUtil.getSpecByName(self, "fillUnit");
	local specSowingMachine = RealisticSeederUtil.getSpecByName(self, "sowingMachine");
	local specDischargeable = RealisticSeederUtil.getSpecByName(self, "dischargeable");
	local dischargeNode = nil;
	
	if specDischargeable ~= nil then
		dischargeNode = specDischargeable.currentRaycastDischargeNode;
	end;
	
	if specFillUnit.unloading ~= nil then
		local isActive = false;
		local isSeedFillType, isFertilizerFillType, isLimeFillType, allowDefaultUnload, _, _ = RealisticSeederUtil.getLoadedFillTypes(self, self:getIsActiveForInput(), specFillUnit, #specFillUnit.fillUnits);

		if specSowingMachine == nil then
			allowDefaultUnload = not isSeedFillType and not isFertilizerFillType and not isLimeFillType;
		end;

		for fillUnitFillTypeIndex, fillUnit in ipairs(self:getFillUnits()) do
			local fillLevel = self:getFillUnitFillLevel(fillUnitFillTypeIndex);
			local fillTypeIndex = self:getFillUnitFillType(fillUnitFillTypeIndex);
			local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex);

			if fillUnit.canBeUnloaded and fillLevel > 0 and fillType.palletFilename ~= nil and not specFillUnit.fillTrigger.isFilling and dischargeNode == nil then
				isActive = allowDefaultUnload;

				break;
			end;
		end;

		g_inputBinding:setActionEventActive(specFillUnit.unloadActionEventId, isActive);
	end;
end;

FillUnit.updateUnloadActionDisplay = RealisticSeeder.updateUnloadActionDisplay;

function RealisticSeeder:updateFillLevelFrames(superFunc)
	local _, yOffset = self:getPosition();

	for fillLevelBufferNumber = 1, #self.fillLevelBuffer do
		local fillLevelInformation = self.fillLevelBuffer[fillLevelBufferNumber];

		if self.fillTypeFrames[fillLevelInformation.fillType] ~= nil then 
			local value = 0;

			if fillLevelInformation.capacity > 0 then
				value = fillLevelInformation.fillLevel / fillLevelInformation.capacity;
			end;

			local frame = self.fillTypeFrames[fillLevelInformation.fillType];

			frame:setVisible(true);

			local fillBar = self.fillTypeLevelBars[fillLevelInformation.fillType];
			local _, yScale = fillBar:getScale();

			fillBar:setScale(value * self.uiScale, yScale);

			local posX, posY = frame:getPosition();

			frame:setPosition(posX, yOffset);

			local fillText = string.format("%d (%d%%)", MathUtil.round(fillLevelInformation.fillLevel), 100 * value);

			self.fillLevelTextBuffer[#self.fillLevelTextBuffer + 1] = fillText;

			if fillLevelInformation.fillType ~= FillType.UNKNOWN then
				local fillTypeText = g_fillTypeManager:getFillTypeByIndex(fillLevelInformation.fillType).title;
				
				self.fillTypeTextBuffer[#self.fillLevelTextBuffer] = fillTypeText;
			end

			yOffset = yOffset + self.frameHeight;
		else
			--## fix, create fill type frame when nil, instead of spamming the log file full... Thanks again to GIANTS for this issue.
			local posX, posY = self:getPosition();
			local frame = self:createFillTypeFrame(self.hudAtlasPath, posX, posY, g_fillTypeManager:getFillTypeByIndex(fillLevelInformation.fillType));

			self.fillTypeFrames[fillLevelInformation.fillType] = frame;

			frame:setScale(self.uiScale, self.uiScale);

			self:addChild(frame);
		end;
	end;
end;

FillLevelsDisplay.updateFillLevelFrames = Utils.overwrittenFunction(FillLevelsDisplay.updateFillLevelFrames, RealisticSeeder.updateFillLevelFrames);

function RealisticSeeder:updateFillLevelBuffers(superFunc)
	local function clearTable(table)
		for tableNumber = #table, 1, -1 do
			table[tableNumber] = nil;
		end;
	end;
	
	clearTable(self.fillLevelTextBuffer);
    clearTable(self.fillLevelBuffer);
	clearTable(self.fillTypeTextBuffer);
	
    for _, frame in pairs(self.fillTypeFrames) do
		if frame ~= nil and frame.setVisible ~= nil then	
			frame:setVisible(false);
		end;
	end;
	
    self.vehicle:getFillLevelInformation(self.fillLevelBuffer);
end;

FillLevelsDisplay.updateFillLevelBuffers = Utils.overwrittenFunction(FillLevelsDisplay.updateFillLevelBuffers, RealisticSeeder.updateFillLevelBuffers);

--## Hook loadSpecValueFillTypes function from FillUnit.lua to add potatoes, fertilizer and seeds, if it necessary, in the shop view
FillUnit.loadSpecValueFillTypes = function(xmlFile, customEnvironment)
	local fillTypeNames, fillTypeCategoryNames, fillTypes, fruitTypeNames = nil;
	local rootName = getXMLRootName(xmlFile);
	local fillUnitConfigurationNumber = 0;
	local holmerPotatoReadyMod = RealisticSeederUtil.getModByTitle("Holmer Potato Ready");

	while true do
		local fillUnitConfigurationKey = string.format(rootName .. ".fillUnit.fillUnitConfigurations.fillUnitConfiguration(%d)", fillUnitConfigurationNumber);

		if not hasXMLProperty(xmlFile, fillUnitConfigurationKey) then
			break;
		end;

		local fillUnitNumber = 0;

		while true do
			local fillUnitKey = string.format(fillUnitConfigurationKey .. ".fillUnits.fillUnit(%d)", fillUnitNumber);

			if not hasXMLProperty(xmlFile, fillUnitKey) then
				break;
			end;

			local showInShop = getXMLBool(xmlFile, fillUnitKey .. "#showInShop");
			local capacity = getXMLFloat(xmlFile, fillUnitKey .. "#capacity");

			if (showInShop == nil or showInShop) and (capacity == nil or capacity > 0) then
				local currentFillTypes = getXMLString(xmlFile, fillUnitKey .. "#fillTypes");

				if currentFillTypes ~= nil then
					if holmerPotatoReadyMod ~= "" then
						if string.find(currentFillTypes, "sugarBeet") and not string.find(currentFillTypes, "potato") then
							currentFillTypes = currentFillTypes .. " potato"; --## add potato to the shop view
						end;
					end;

					if fillTypeNames == nil then
						fillTypeNames = currentFillTypes;
					else
						fillTypeNames = fillTypeNames .. " " .. currentFillTypes;
					end;
				end;

				local currentFillTypeCategories = getXMLString(xmlFile, fillUnitKey .. "#fillTypeCategories");
				
				if RealisticSeederUtil ~= nil and RealisticSeederUtil.foundFunctionAdditionalTank(xmlFile) and RealisticSeederUtil.getIsValidStoreCategory(getXMLString(xmlFile, "vehicle.storeData.category")) then
					currentFillTypeCategories = nil;

					fillTypeNames = "fertilizer seeds";
				end;

				if currentFillTypeCategories ~= nil then
					if fillTypeCategoryNames == nil then
						fillTypeCategoryNames = currentFillTypeCategories;
					else
						fillTypeCategoryNames = fillTypeCategoryNames .. " " .. currentFillTypeCategories;
					end;
				end;
			end;

			fillUnitNumber = fillUnitNumber + 1;
		end;

		fillUnitConfigurationNumber = fillUnitConfigurationNumber + 1;
	end;

	if fillTypeNames == nil then
		fillTypeNames = getXMLString(xmlFile, rootName .. ".fillTypes");
	end;

	fillTypeNames = Utils.getNoNil(getXMLString(xmlFile, rootName .. ".storeData.specs.fillTypes"), fillTypeNames);

	if fillTypeNames ~= nil then
		fillTypeCategoryNames = nil;
	end;

	if fillTypeCategoryNames == nil then
		fillTypeCategoryNames = getXMLString(xmlFile, rootName .. ".fillTypeCategories");
	end;

	if fillTypes == nil then
		fruitTypeNames = getXMLString(xmlFile, rootName .. ".fruitTypes");
	end;

	if fillTypes == nil then
		fruitTypeNames = getXMLString(xmlFile, rootName .. ".cutter#fruitTypes");

		if fruitTypeNames ~= nil and holmerPotatoReadyMod ~= "" then
			if string.find(fruitTypeNames, "sugarBeet") and not string.find(fruitTypeNames, "potato") then
				fruitTypeNames = fruitTypeNames .. " potato"; --## add potato to the shop view
			end;
		end;
	end;

	local fruitTypeCategoryNames = getXMLString(xmlFile, rootName .. ".cutter#fruitTypeCategories");
	local useWindrowed = Utils.getNoNil(getXMLBool(xmlFile, rootName .. ".cutter#useWindrowed"), false);

	fillTypeCategoryNames = Utils.getNoNil(getXMLString(xmlFile, rootName .. ".storeData.specs.fillTypeCategories"), fillTypeCategoryNames);

	return {
		categoryNames = fillTypeCategoryNames,
		fillTypeNames = fillTypeNames,
		fruitTypeNames = fruitTypeNames,
		fruitTypeCategoryNames = fruitTypeCategoryNames,
		useWindrowed = useWindrowed
	};
end;

----------------------------------------------------------------------------------------------------------------------------------
---------------------------------------RealisticSeederUnloadFillUnitsToPalletsEvent-----------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------

RealisticSeederUnloadFillUnitsToPalletsEvent = {};

local RealisticSeederUnloadFillUnitsToPalletsEvent_mt = Class(RealisticSeederUnloadFillUnitsToPalletsEvent, Event);

InitEventClass(RealisticSeederUnloadFillUnitsToPalletsEvent, "RealisticSeederUnloadFillUnitsToPalletsEvent");

function RealisticSeederUnloadFillUnitsToPalletsEvent:emptyNew()
	local self = Event:new(RealisticSeederUnloadFillUnitsToPalletsEvent_mt);

	return self;
end;

function RealisticSeederUnloadFillUnitsToPalletsEvent:new(object, unloadFertilizer, unloadLime)
	local self = RealisticSeederUnloadFillUnitsToPalletsEvent:emptyNew();

	self.object = object;
	self.unloadFertilizer = unloadFertilizer;
	self.unloadLime = unloadLime;

	return self;
end;

function RealisticSeederUnloadFillUnitsToPalletsEvent:newServerToClient()
	local self = RealisticSeederUnloadFillUnitsToPalletsEvent:emptyNew();

	return self;
end;

function RealisticSeederUnloadFillUnitsToPalletsEvent:readStream(streamId, connection)
	if not connection:getIsServer() then
		self.object = NetworkUtil.readNodeObject(streamId);

		self.unloadFertilizer = streamReadBool(streamId);
		self.unloadLime = streamReadBool(streamId);
	end;

	self:run(connection);
end;

function RealisticSeederUnloadFillUnitsToPalletsEvent:writeStream(streamId, connection)
	if connection:getIsServer() then
		NetworkUtil.writeNodeObject(streamId, self.object);

		streamWriteBool(streamId, self.unloadFertilizer);
		streamWriteBool(streamId, self.unloadLime);
	end;
end;

function RealisticSeederUnloadFillUnitsToPalletsEvent:run(connection)
	if not connection:getIsServer() then
		if self.object ~= nil then
			local success = self.object:unloadFillUnitsToPallets(false, self.unloadFertilizer, self.unloadLime);

			if not success then
				connection:sendEvent(RealisticSeederUnloadFillUnitsToPalletsEvent:newServerToClient());
			end;
		end;
	else
		g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_INFO, g_i18n:getText("fillUnit_unload_nospace"));
	end;
end;

--------------------------------------------------------------------------------------------------------------------------
---------------------------------------RealisticSeederSwitchUnloadModeEvent-----------------------------------------------
--------------------------------------------------------------------------------------------------------------------------

RealisticSeederSwitchUnloadModeEvent = {};

local RealisticSeederSwitchUnloadModeEvent_mt = Class(RealisticSeederSwitchUnloadModeEvent, Event);

InitEventClass(RealisticSeederSwitchUnloadModeEvent, "RealisticSeederSwitchUnloadModeEvent");

function RealisticSeederSwitchUnloadModeEvent:emptyNew()
	local self = Event:new(RealisticSeederSwitchUnloadModeEvent_mt);
    
	return self;
end;

function RealisticSeederSwitchUnloadModeEvent:new(vehicle, isUnloadingToPallets)
	local self = RealisticSeederSwitchUnloadModeEvent:emptyNew();
	
	self.vehicle = vehicle;
	self.isUnloadingToPallets = isUnloadingToPallets;

	return self;
end;

function RealisticSeederSwitchUnloadModeEvent:readStream(streamId, connection)
	self.vehicle = NetworkUtil.readNodeObject(streamId);
	self.isUnloadingToPallets = streamReadBool(streamId);
	
	self:run(connection);
end;

function RealisticSeederSwitchUnloadModeEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.vehicle);

	streamWriteBool(streamId, self.isUnloadingToPallets);
end;

function RealisticSeederSwitchUnloadModeEvent:run(connection)
	if not connection:getIsServer() then
		g_server:broadcastEvent(RealisticSeederSwitchUnloadModeEvent:new(self.vehicle, self.isUnloadingToPallets), nil, connection, self.vehicle);
	end;
	
    if self.vehicle ~= nil then
        self.vehicle:switchUnloadMode(self.isUnloadingToPallets, true);
	end;
end;

function RealisticSeederSwitchUnloadModeEvent.sendEvent(vehicle, isUnloadingToPallets, noEventSend)
	if not noEventSend then
		if g_server ~= nil then
			g_server:broadcastEvent(RealisticSeederSwitchUnloadModeEvent:new(vehicle, isUnloadingToPallets), nil, nil, vehicle);
		else
			g_client:getServerConnection():sendEvent(RealisticSeederSwitchUnloadModeEvent:new(vehicle, isUnloadingToPallets));
		end;
	end;
end;