--[[
RealisticSeederUtil

Utility for my Realistic Seeder

Author:		Ifko[nator]
Datum:		31.07.2021
Version:	v1.8

History:	v1.0 @12.09.2020 - initial implementation in FS 19
			----------------------------------------------------------------------------------------------------
			v1.1 @15.02.2021 - reworkt changeMaterial function
			----------------------------------------------------------------------------------------------------
			v1.2 @06.06.2021 - added "getFertilizerFillUnitIndex" and "getSeedFillTypes" function
			----------------------------------------------------------------------------------------------------
			v1.3 @13.06.2021 - minior adjustments
			----------------------------------------------------------------------------------------------------
			v1.4 @15.06.2021 - added "getIsValidPalletName" function
			----------------------------------------------------------------------------------------------------
			v1.5 @21.06.2021 - added "getLoadedFillTypes" and removed "getFertilizerFillUnitIndex" function
			----------------------------------------------------------------------------------------------------
			v1.6 @23.06.2021 - added "fixFillTypeName" function
			----------------------------------------------------------------------------------------------------
			v1.7 @07.07.2021 - added "getIsUnSupportetValue" function and "unSupportetHofBergmannVersions" table
			----------------------------------------------------------------------------------------------------
			v1.8 @31.07.2021 - added "getIsValidStoreCategoryName" function
]]

RealisticSeederUtil = {};
RealisticSeederUtil.currentModName = g_currentModName;
RealisticSeederUtil.languageShort = g_languageShort;
RealisticSeederUtil.currentModDirectory = g_currentModDirectory;
RealisticSeederUtil.categoryPrefix = "REALISTICSEEDER";
RealisticSeederUtil.maxPalletsAmount = 10;

RealisticSeederUtil.validStoreCategories = {
    "seeders",
    "planters",
	"potatoHarvesting",
	"weeders",
	"discHarrows",
	"cultivators",
	"kvernelandPack"
};

RealisticSeederUtil.unSupportetHofBergmannVersions = {
	"1.0.0.0",
	"1.0.0.2",
	"1.0.0.3",
	"1.0.0.4",
	"1.0.0.5",
	"1.0.0.6",
	"1.0.0.61",
	"1.0.0.62",
	"1.0.0.7",
	"1.0.0.8"
};

RealisticSeederUtil.supportetCategoryNames = {
	RealisticSeederUtil.categoryPrefix .. "PALLETS",
	RealisticSeederUtil.categoryPrefix .. "BIGBAGS",
	RealisticSeederUtil.categoryPrefix .. "BAGS"
};

RealisticSeederUtil.modDesc = loadXMLFile("modDesc", RealisticSeederUtil.currentModDirectory .. "modDesc.xml");
RealisticSeederUtil.fillTypesPath = RealisticSeederUtil.currentModDirectory .. getXMLString(RealisticSeederUtil.modDesc, "modDesc.fillTypes#filename");
	
if fileExists(RealisticSeederUtil.fillTypesPath) then
	RealisticSeederUtil.fillTypesXML = loadXMLFile("map", RealisticSeederUtil.fillTypesPath);
end;

getfenv(0)["RealisticSeederUtil"] = RealisticSeederUtil;

function RealisticSeederUtil.getDebugPriority(xmlFile, path)
	return Utils.getNoNil(getXMLInt(xmlFile, path), 0);
end;

function RealisticSeederUtil.printError(errorMessage, isWarning, isInfo, luaName)
	local prefix = "::ERROR:: ";
	
	if isWarning then
		prefix = "::WARNING:: ";
	elseif isInfo then
		prefix = "::INFO:: ";
	end;
	
	print(prefix .. "from the " .. luaName .. ".lua: " .. tostring(errorMessage));
end;

function RealisticSeederUtil.printDebug(debugMessage, priority, addString, luaName)
	if priority >= 1 then
		local prefix = "";
		
		if addString then
			prefix = "::DEBUG:: from the " .. luaName .. ".lua: ";
		end;
		
		print(prefix .. tostring(debugMessage));
	end;
end;

function RealisticSeederUtil.getSpecByName(self, specName, currentModName)
    local spec = self["spec_" .. Utils.getNoNil(currentModName, RealisticSeederUtil.currentModName) .. "." .. specName];

	if spec ~= nil then
        return spec;
    end;

    return self["spec_" .. specName];
end;

function RealisticSeederUtil.getIsValidStoreCategory(storeCategory)
    for _, validStoreCategory in pairs(RealisticSeederUtil.validStoreCategories) do
        if storeCategory == validStoreCategory then
            return true;
        end;
    end;

    return false;
end;

function RealisticSeederUtil.getIsValidStoreCategoryName(storeCategoryName)
    for _, supportetCategoryName in pairs(RealisticSeederUtil.supportetCategoryNames) do
        if storeCategoryName == supportetCategoryName then
            return true;
        end;
    end;

    return false;
end;

function RealisticSeederUtil.getIsValidPalletName(palletName)
	return (not string.find(palletName, "PALLETFORK") and not string.find(palletName, "TRIGGER")) and (string.find(palletName, "PALLET") or string.find(palletName, "BIGBAG") or string.find(palletName, "TANK"));
end;

function RealisticSeederUtil.foundFunctionAdditionalTank(xmlFile)
    local foundFunctionAdditionalTank = false;
	local additionalTankNumber = 0;

	while true do
		local functionAditionalTankKey = "vehicle.storeData.functions.function(" .. additionalTankNumber .. ")";
		
		if not hasXMLProperty(xmlFile, functionAditionalTankKey) then
			break;
		end;

		if getXMLString(xmlFile, functionAditionalTankKey) == "$l10n_function_additionalTank" then
			foundFunctionAdditionalTank = true;

			break;
		end;

		additionalTankNumber = additionalTankNumber + 1;
    end;
    
    return foundFunctionAdditionalTank;
end;

function RealisticSeederUtil.getSeedFillUnitIndex(xmlFile, object, specFillUnit)
	local seedFillUnitIndex = 1;
	
	local baseKey = string.format("vehicle.fillUnit.fillUnitConfigurations.fillUnitConfiguration(%d).fillUnits", Utils.getNoNil(object.configurations["fillUnit"], 1) - 1);

	local fillUnitNumber = 0;

	while true do
		local fillUnitKey = string.format("%s.fillUnit(%d)", baseKey, fillUnitNumber);
		
        if not hasXMLProperty(xmlFile, fillUnitKey) then
            break;
		end;

		for fillTypeIndex, _ in pairs(specFillUnit.fillUnits[fillUnitNumber + 1].supportedFillTypes) do
			if string.find(g_fillTypeManager:getFillTypeByIndex(fillTypeIndex).name, "SEED") then
				seedFillUnitIndex = fillUnitNumber + 1;

				break;
			end;
		end;

		fillUnitNumber = fillUnitNumber + 1;
    end;
    
    return seedFillUnitIndex;
end;

function RealisticSeederUtil.getLoadedFillTypes(object, getIsActiveForInput, specFillUnit, numFillUnits)
	local isSeedFillType, isFertilizerFillType, isLimeFillType, allowDefaultUnload, fillUnitFillType, seedFillTypeIndex = false, false, false, false, g_fillTypeManager:getFillTypeByIndex(FillType.UNKNOWN), 0;

	if getIsActiveForInput then
		for fillUnitNumber = 1, numFillUnits do
			local fillUnit = specFillUnit.fillUnits[fillUnitNumber];

			if fillUnit ~= nil and fillUnit.showOnHud then
				fillUnitFillType = g_fillTypeManager:getFillTypeByIndex(object:getFillUnitFillType(fillUnitNumber));

				if fillUnitFillType ~= nil then
					if string.find(fillUnitFillType.name, "SEED") then
						isSeedFillType = true;

						seedFillTypeIndex = object:getFillUnitFillType(fillUnitNumber);
					end;
					
					if fillUnitFillType.name == "FERTILIZER" then
						isFertilizerFillType = true;
					end;
					
					if fillUnitFillType.name == "LIME" then
						isLimeFillType = true;
					end;
					
					if fillUnitFillType.name == "HERBICIDE" or fillUnitFillType.name == "LIQUIDFERTILIZER" then
						allowDefaultUnload = true;
					end;
				end;
			end;
		end;
	end;

	return isSeedFillType, isFertilizerFillType, isLimeFillType, allowDefaultUnload, fillUnitFillType, seedFillTypeIndex;
end;

function RealisticSeederUtil.fixFillTypeName(fillTypeName)
	if fillTypeName == "seed_sugarbeet" then
		fillTypeName = "seed_sugarBeet";
	elseif fillTypeName == "seed_oilseedradish" then
		fillTypeName = "seed_oilseedRadish";
	elseif fillTypeName == "seed_horsegrass" then
		fillTypeName = "seed_horseGrass";
	elseif fillTypeName == "seed_pasturegrass" then
		fillTypeName = "seed_pastureGrass";
	end;

	return fillTypeName;
end;

function RealisticSeederUtil.changeMaterial(object, materialNode, translateNode)
	if object.components == nil then
		local newMaterial = getMaterial(translateNode, 0);
		local numChildren = getNumOfChildren(materialNode);

		if numChildren > 0 then
			for childNumber = 0, numChildren - 1 do
				local oldMaterial = getMaterial(getChildAt(materialNode, childNumber), 0);

				ConfigurationUtil.replaceMaterialRec(object, getChildAt(materialNode, childNumber), oldMaterial, newMaterial);
			end;
		else
			local oldMaterial = getMaterial(materialNode, 0);

			ConfigurationUtil.replaceMaterialRec(object, materialNode, oldMaterial, newMaterial);
		end;
	else
		local oldMaterial = getMaterial(materialNode, 0);
		local newMaterial = getMaterial(translateNode, 0);
	
		for _, component in pairs(object.components) do
			ConfigurationUtil.replaceMaterialRec(object, component.node, oldMaterial, newMaterial);
		end;
	end;
end;

function RealisticSeederUtil.getCanFruitBePlanted(fillType, fruitType)
	local canFruitBePlanted = false;

    if g_currentMission.fruits ~= nil then
        if fruitType ~= nil then
			if g_currentMission.fruits[fruitType] ~= nil
				and g_currentMission.fruits[fruitType].id ~= nil 
				and g_currentMission.fruits[fruitType].id ~= 0 
			then
				canFruitBePlanted = true;
			end;
		else
			if fillType ~= nil then
				local seedsFruitType = g_fruitTypeManager:getFruitTypeByName(string.sub(g_fillTypeManager:getFillTypeByIndex(fillType).name, 6));
				
				if seedsFruitType ~= nil 
					and g_currentMission.fruits[seedsFruitType.index] ~= nil 
					and g_currentMission.fruits[seedsFruitType.index].id ~= nil 
					and g_currentMission.fruits[seedsFruitType.index].id ~= 0 
				then
					canFruitBePlanted = true;
				end;
			end;
		end;
	end;

	return canFruitBePlanted;
end;

function RealisticSeederUtil.getL10nText(text)
	if text:sub(1, 6) == "$l10n_" then
		text = g_i18n:getText(text:sub(7));
	elseif g_i18n:hasText(text) then
		text = g_i18n:getText(text);
	end;

	return text;
end;

function RealisticSeederUtil.getMultipleText(typeDesc)
	local text = RealisticSeederUtil.getL10nText(typeDesc);
	
	if typeDesc == "$l10n_typeDesc_pallet" then	
		text = RealisticSeederUtil.getL10nText("category_pallets");
	elseif typeDesc == "$l10n_typeDesc_bag" then	
		text = string.sub(RealisticSeederUtil.getL10nText("storeTitle_realisticSeederBags"), 18);
	elseif typeDesc == "Bigbag" then	
		text = typeDesc .. "s";
	end;

	return text;
end;

function RealisticSeederUtil.getSeedFillTypes()
	local fillTypesXML = loadXMLFile("map", RealisticSeederUtil.currentModDirectory .. getXMLString(RealisticSeederUtil.modDesc, "modDesc.fillTypes#filename"));

	local fillTypes = nil;
	local fillTypeNames = "";
    local fillTypeNumber = 0;

    while true do
		local fillTypeKey = "map.fillTypes.fillType(" .. tostring(fillTypeNumber) .. ")";

		if not hasXMLProperty(fillTypesXML, fillTypeKey) then
			break;
		end;

        local fillType = Utils.getNoNil(getXMLString(fillTypesXML, fillTypeKey .. "#name"), "");

        if fillType ~= "" then
			if fillTypeNames ~= "" then 
				fillType = " " .. fillType;
			end;

			fillTypeNames = fillTypeNames .. fillType;
		end;
        
        fillTypeNumber = fillTypeNumber + 1;
    end;

	fillTypes = g_fillTypeManager:getFillTypesByNames(fillTypeNames, "Warning: invalid fillType '%s'.");

	return fillTypes;
end;

function RealisticSeederUtil.getModByTitle(germanTitle, englishTitle)
	local mod, modName, modDir = nil, "", "";

	if englishTitle == nil then
		englishTitle = germanTitle;
	end;

	if germanTitle == nil then
		return nil, "", "";
	end;
	
	for _, mod in pairs(g_modManager.mods) do
		if mod.title == germanTitle or mod.title == englishTitle then		
			if g_modIsLoaded[tostring(mod.modName)] then	
				mod = mod;
				modName = mod.modName;
				modDir = mod.modDir;

				break;
			end;
		end;
	end;

	return mod, modName, modDir;
end;

function RealisticSeederUtil.getIsUnSupportetValue(unSupportetValues, unSupportetValue)
	for unSupportetValueNumber = 1, #unSupportetValues do
		if unSupportetValues[unSupportetValueNumber] == unSupportetValue then
			return true;
		end;
	end;
	
	return false;
end;